"""
Set the logging level for pyiron
"""

import logging
from types import MethodType


def set_logging_level(self, level, channel=None):
    """
    Set level for logger

    Args:
        level (str): 'DEBUG, INFO, WARN'
        channel (int): 0: file_log, 1: stream, None: both
    """

    if channel:
        self.handlers[channel].setLevel(level)
    else:
        self.handlers[0].setLevel(level)
        self.handlers[1].setLevel(level)


def setup_logger():
    """
    Setup logger - logs are written to pyiron.log

    Returns:
        logging.getLogger: Logger
    """
    logger = logging.getLogger("pyiron_log")
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    ch = logging.StreamHandler()
    ch.setLevel(logging.WARN)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    try:
        fh = logging.FileHandler("pyiron.log")
    except PermissionError:
        pass
    else:
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


logger = setup_logger()
logger.set_logging_level = MethodType(set_logging_level, logger)



import contextlib
import os


@contextlib.contextmanager
def TemporaryEnvironment(**kwargs):
    """
    Context manager for temporarily setting environment variables.

    Takes any number of keyword arguments where the key is the environment
    variable to set and the value is the value to set it to. For the duration
    of the context, the environment variables are set as per the provided arguments.
    The original environment setting is restored once the context is exited,
    even if an exception is raised within the context.

    Non-string values are coerced with `str()`.

    Can also be used as a function decorator.

    Examples:

    >>> import os
    >>> with TemporaryEnvironment(PATH='/tmp', HOME='/', USER='foobar'):
    ...     print(os.getenv('PATH'))  # Outputs: /tmp
    ...     print(os.getenv('HOME'))  # Outputs: /
    ...     print(os.getenv('USER'))  # Outputs: foobar
    /tmp
    /
    foobar
    """
    old_vars = {}
    for k, v in kwargs.items():
        with contextlib.suppress(KeyError):
            old_vars[k] = os.environ[k]
        os.environ[k] = str(v)
    try:
        yield
    finally:
        for k in kwargs:
            if k in old_vars:
                os.environ[k] = old_vars[k]
            else:
                del os.environ[k]



from __future__ import annotations

import tarfile
import uuid
from pathlib import Path
from typing import cast


def delete_files_and_directories_recursively(path):
    if not path.exists():
        return
    for item in path.rglob("*"):
        if item.is_file():
            item.unlink()
        else:
            delete_files_and_directories_recursively(item)
    path.rmdir()


def categorize_folder_items(folder_path):
    if not folder_path.is_dir():
        return {}
    types = [
        "dir",
        "file",
        "mount",
        "symlink",
        "block_device",
        "char_device",
        "fifo",
        "socket",
    ]
    results = {t: [] for t in types}

    for item in folder_path.iterdir():
        for tt in types:
            try:
                if getattr(item, f"is_{tt}")():
                    results[tt].append(str(item))
            except NotImplementedError:
                pass
    return results


class DirectoryObject:
    """
    A class to represent a directory object that can be created, deleted,
    and managed. It can also compress and decompress its contents.
    It supports unique directory generation and can be protected from deletion
    on garbage collection.
    """

    def __init__(
        self,
        directory: str | Path | DirectoryObject = ".",
        generate_unique_directory: bool | None = None,
        protected: bool | None = None,
    ):
        """
        Initialize a DirectoryObject.

        Args:
            directory (str | Path | DirectoryObject): The directory path or
                DirectoryObject instance.
            generate_unique_directory (bool | None): If True, generates a unique
                directory name, otherwise it still generates a unique name if
                the directory is "." and this parameter is None.
            protected (bool | None): If True, prevents deletion of the
                directory object on garbage collection. If None, it defaults to
                True if the directory already exists.
        """
        if isinstance(directory, str):
            path = Path(directory)
        elif isinstance(directory, Path):
            path = directory
        elif isinstance(directory, DirectoryObject):
            path = directory.path
        if (
            directory == "." and generate_unique_directory is None
        ) or generate_unique_directory:
            path = path / f"data_{uuid.uuid4().hex}"
        if protected is None:
            protected = path.exists()
        self._protected = protected
        self.path: Path = path
        self.create()

    def __getstate__(self):
        self._protected = True
        return self.path.__getstate__()

    def __del__(self):
        if not self._protected:
            self.delete(only_if_empty=False)

    def create(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def delete(self, only_if_empty: bool = False):
        if self.is_empty() or not only_if_empty:
            delete_files_and_directories_recursively(self.path)

    def list_content(self):
        return categorize_folder_items(self.path)

    def __len__(self):
        return sum([len(cc) for cc in self.list_content().values()])

    def __repr__(self):
        return f"DirectoryObject(directory='{self.path}')\n{self.list_content()}"

    def get_path(self, file_name):
        return self.path / file_name

    def file_exists(self, file_name):
        return self.get_path(file_name).is_file()

    def write(self, file_name, content, mode="w"):
        with self.get_path(file_name).open(mode=mode) as f:
            f.write(content)

    def create_subdirectory(self, path):
        return DirectoryObject(self.path / path)

    def is_empty(self) -> bool:
        return len(self) == 0

    def remove_files(self, *files: str):
        for file in files:
            path = self.get_path(file)
            if path.is_file():
                path.unlink()

    def compress(self, exclude_files: list[str | Path] | None = None):
        directory = self.path.resolve()
        output_tar_path = directory.with_suffix(".tar.gz")
        if output_tar_path.exists():
            return
        if exclude_files is None:
            exclude_files = []
        else:
            exclude_files = [Path(f) for f in exclude_files]
        exclude_set = {
            f.resolve() if f.is_absolute() else (directory / f).resolve()
            for f in cast(list[Path], exclude_files)
        }
        files_to_delete = []
        with tarfile.open(output_tar_path, "w:gz") as tar:
            for file in directory.rglob("*"):
                if file.is_file() and file.resolve() not in exclude_set:
                    arcname = file.relative_to(directory)
                    tar.add(file, arcname=arcname)
                    files_to_delete.append(file)
        for file in files_to_delete:
            file.unlink()

    def decompress(self):
        directory = self.path.resolve()
        tar_path = directory.with_suffix(".tar.gz")
        if not tar_path.exists():
            return
        with tarfile.open(tar_path, "r:gz") as tar:
            tar.extractall(path=directory, filter="fully_trusted")
        tar_path.unlink()



from __future__ import annotations

import time
import warnings
from collections.abc import Callable
from itertools import count
from typing import Any, TypeVar

T = TypeVar("T")


def retry(
    func: Callable[[], T],
    error: type[Exception] | tuple[type[Exception], ...],
    msg: str,
    at_most: int | None = None,
    delay: float = 1.0,
    delay_factor: float = 1.0,
    log: bool | Any = True,
) -> T:
    """
    Try to call `func` until it no longer raises `error`.

    Any other exception besides `error` is still raised.

    Args:
        func (callable): function to call, should take no arguments
        error (Exception or tuple thereof): any exceptions to be caught
        msg (str): messing to be written to the log if `error` occurs.
        at_most (int, optional): retry at most this many times, None means retry
                                forever
        delay (float): time to wait between retries in seconds
        delay_factor (float): multiply `delay` between retries by this factor
        logger (bool|object): Whether to pass a message to `warnings.warn` on each
            retry. (Default is True.) Optionally, an object with a :meth:`warn` method
            can be passed and the message will be sent there instead
            (e.g. `pyiron_snippets.logger.logger`).

    Raises:
        `error`: if `at_most` is exceeded the last error is re-raised
        Exception: any exception raised by `func` that does not match `error`

    Returns:
        object: whatever is returned by `func`
    """
    tries = count() if at_most is None else range(at_most)
    for i in tries:
        try:
            return func()
        except error as e:
            warning = f"{msg} Trying again in {delay}s. Tried {i + 1} times so far..."
            if isinstance(log, bool):
                if log:
                    warnings.warn(warning, stacklevel=2)
            else:
                log.warn(warning)
            time.sleep(delay)
            delay *= delay_factor
            # e drops out of the namespace after the except clause ends, so
            # assign it here to a dummy variable so that we can re-raise it
            # in case the error persists
            err = e
    raise err from None



"""
Simple stuff for colors when you don't want matplotlib/seaborn in your dependency stack.
"""


class SeabornColors:
    """
    Hex codes for the ten `seaborn.color_palette()` colors (plus pure white and black),
    recreated to avoid adding an entire dependency.
    """

    blue = "#1f77b4"
    orange = "#ff7f0e"
    green = "#2ca02c"
    red = "#d62728"
    purple = "#9467bd"
    brown = "#8c564b"
    pink = "#e377c2"
    gray = "#7f7f7f"
    olive = "#bcbd22"
    cyan = "#17becf"
    white = "#ffffff"
    black = "#000000"
